/*
  ------------------------------------------------------------------------------

    Copyright © 2005-2012, Tommi T. Järvi
                           firstname . jarvi @ iki . fi (remove spaces)


    This file is part of one of my programs.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2,
    as published by the Free Software Foundation.

    This is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  ------------------------------------------------------------------------------
*/


#include <valarray>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cassert>
#include <string>
#include <vector>
#include <algorithm>
#include <cctype>
#include "mdconstants.h"
#include "global.h"
#include "mathsubs.h"
#ifndef DEF__ATOM_TYPES
#include "atomtypes.h"
#endif

using namespace std;


#ifndef DEF__ATOM_SYSTEM
#define DEF__ATOM_SYSTEM

#include <gsl/gsl_complex.h>  // used by sfactor
#include <gsl/gsl_complex_math.h>  // used by sfactor
#include <gsl/gsl_sf_legendre.h>  // used by q_lm
#include <gsl/gsl_complex.h>  // used by q_lm
#include <gsl/gsl_complex_math.h>  // used by q_lm
#include <gsl/gsl_sf_gamma.h>  // used by q_lm


class Atom_System {

 private:

  // --- basic properties

  int dim;  // dimension of space
  int natoms;  // number of atoms

  // --- other properties

  string comment;

  // --- neighbour list(s)

  int maxneighbours;

 public:

  // --- atoms

  int x0size;  // size of x_i, F, fixx0, dx0
  double *x0;  // coordinates
  double *x1;  // velocities
  double *F;   // forces
  int *fixx0;  // fixed coordinates

  int atypesize;  // size of atype, V, K, block
  int *atype;     // atom types
  double *V;      // potential energies
  double *K;      // kinetic energies
  int *block;     // cluster

  valarray<int> blocknat;  // number of atoms in different blocks, same as the above

  Atom_Types atypes;  // atom types

  // --- properties

  int *pbc;  // periodic boundary conditions (0 no, 1 yes)
  double *box;  // box size for periodic b.c.s and mdlat-files

  double T;  // temperature (K)
  double a;  // lattice constant

  // --- neighbour list(s)

  double *dx0;  // displacements since last nlist update (size x0size)

  int nlistsize;  // size of nlist array
  int *nlist;  // Verlet neighbour list
  double nlist_cut;  // cutoff of above
  double nlist_potcut;  // cutoff of potential

  vector< valarray<int> > symmnlist;  // symmetric neighbour list: symmnlist[i][j]: index of neighbour j of atom i

  int cells;  // number of linked list cells
  double *lsubcell;  // length of linked list subcell
  int *nsubcell;  // number of subcells in different directions
  int *linkedlist_head;  // head-array (0, ..., cells-1)
  int *linkedlist_list;  // list-array (0, ..., natoms-1 => size atypesize)

  // --- operators

  ~Atom_System();
  Atom_System();
  Atom_System(const Atom_System &syst);
  Atom_System& operator=(const Atom_System &syst);
  void delete_arrays();  // deletes all arrays

  // --- setting stuff

  void moreatoms(int n);  // make room for n more atoms
  void reset(int n);  // reset system size to n atoms
  void init_with_ucell(int crystal, const valarray<int> &size, valarray<int> atom_types, vector<string> atom_names, int offsetmode, int mode = 0);
  void init_with_ucell(int crystal, const valarray<int> &size, const valarray<double> &alat, valarray<int> atom_types, vector<string> atom_names, int offsetmode, int mode = 0);
  void read_unit_cell_and_init(string filename, valarray<int> size, double scalef, int readmode);
  void init_unit_cell(const valarray<int> &size, const valarray<double> &basis_vectors, const valarray<double> &basis_coordinates, valarray<int> basis_atypes, vector<string> basis_anames, int mode);
  void init_FCC_twin(const string atname, const valarray<int> &size, int twinn);
  void random_types(valarray<int> rtypes, vector<string> rnames, double pfirst);
  void random_alloy(valarray<int> rtypes, vector<string> rnames, valarray<double> conc);
  void makewulff(double cut100, double cut111);
  void createmti(double clayers, string crname, int crtype);
  void createrandomcl(int nat, double ra, double rb, string crname, int crtype);

  void set_lattice_constant(double value);
  void setcomment(string com);
  void setpbc(valarray<int> cond);
  void setbox(valarray<double> size);
  void set_temperature(double temp, double tdebye);
  void add_atom(valarray<double> pos, string name, int extrasize);
  void add_atom(const Atom_System &other, int at);
  void join(const Atom_System &other);

  // --- getting stuff

  double dr(int i, int j) const;  // distance of atoms i, j
  double dr(int i, const Atom_System &system, int j) const;  // distance of atoms i, j, with j in a different system
  double dr(int at, valarray<double> point) const;  // distance between atom at and a point in space
  double angle(int i, int j, int k) const;  // angle between atoms ijk in radians
  valarray<double> dri(int i, int j) const;  // return vector x0[i] - x0[j]
  valarray<double> dri(int i, int j, double &dr) const;  // return vector x0[i] - x0[j] and i-j distance
  valarray<double> ri(int i) const;  // return position of atom i
  int atoms() const;
  int dimension() const;
  string get_comment() {return comment;}
  double Vtot();
  int find_closest_atom(const valarray<double> &pos, double &r);
  int find_closest_atom(int i, double &r);
  void cluster(double cutoff);
  void leaveblock(int bl);
  void removeblock(int bl);
  void removehalfspace(valarray<double> base, valarray<double> dir);
  int largestblock();
  bool comment_time(double &time);  // get time from comment line, if possible
  double extrcoord(int c, int d);  // get min (d=0) or max (1) position with respect to coordinate c ([0,2])

  // --- moving / changing stuff

  void moveatom(int i, const valarray<double> &dr, bool movefixed = false);
  void translate(const valarray<double> &dx0, bool movefixed = false);
  void rotate_euler(double phi, double theta, double psi);
  void rotate_axis(double phi, double theta, double delta);
  void center();
  void centerv();
  void getmoved(double rcut, Atom_System other);
  void removeat(int at, int mode = 1);
  void removetype(int type);
  void removetype(string name);
  void removerand(int n);
  void remove(int dir, bool above, double cut);
  void removeV(double val, double tol);
  void removeK(double ka, double kb);  // remove atoms with kinetic energy (K[i]) between ka and kb
  void scalec(double factor);  // scale x0, dx0 by a factor, nothing else touched
  void addEk(double Eat);  // add kinetic energy
  void strain(vector< valarray<double> > &eps);  // strain system, eps get's replaced by new box vectors

  // --- units
  void unit_l(double factor);  // multiply unit of length by factor
  void unit_t(double factor);  // multiply unit of time by factor

  // --- input / output

  // xyz
  void savexyz(string fname, int xyzmode, bool overwrite = false);
  void savexyz(string fname, int savetype, int xyzmode, bool overwrite = false);  // only save atoms of type savetype
  void readxyz(string fname, int xyzmode, int frame = 1);  // read a frame from an md.movie file
  void readxyzframe(ifstream &afile, int xyzmode);
  void xyzmodes();  // print xyzmodes to stdout

  // Adri van Duin's ReaxFF code
  void readbgf(string fname);
  void readvels(string fname);
  void savevels(string fname);
  void savebgf(string fname, string descr, int mode, bool overwrite = false);
  void savebgf(string fname, string descr, int mode, valarray<double> boxsize, valarray<double> angles, bool overwrite = false);

  // Fraunhofer IWM in-house md code
  void savemdcore(string fname, int mode);

  // VASP
  void saveposcar(string fname, int mode, valarray<int> order);  // save poscar-file (the VASP file format)

  // LAMMPS
  void readlammps_data(string fname, int mode = 1);  // Read LAMMPS data file
  void savelammps(string fname, int mode, bool overwrite, string extra_headers);  // Save LAMMPS data file
  double readlammps_custom_dump_frame(ifstream &afile, bool skip = false);  // Read LAMMPS "dump custom" frame

  // Other
  void save_siesta_fdf(string fname);
  void savepov(string fname, int mode, double min, double max, bool overwrite = false);  
  void report();

  // --- others

  void interpolate(const Atom_System &second, double weight);
  void updateT();
  void angmom(double &angmom, double &f, double &J, double &TcorrtoL);
  double sro(double cut, int atomtype, int neitype, double zmin, double zmax);
  double smartsro(double cut, int typea, int typeb);  // NOTE THE STUPID SOLUTION TO REMOVE ZMIN, ZMAX!
  double localsro(int atomtype, int neitype, double cuta, double cutb);
  double sronei(int i, int neitype, double cuta, double cutb);
  valarray <double> cmvel();
  double cmekin();
  double cmcoord(valarray<double> &centre);
  double rgyr();  // radius of gyration (correct?)
  double sfactor(valarray<int> n);
  double sfactor(valarray<int> n, valarray<int> incl);
  double sfactor_znorm(double zlimit, valarray<int> n);
  double paircorr(double r, double dr);  // pair correlation function at r with shell thickness dr
  gsl_complex q_lm(int at, int l, int m);
  double q_l(int at, int l);
  double Q_l(int l);
  gsl_complex Q_lm(int l, int m);
  double q_correlation(int i, int j, int l);

  gsl_complex y_lm(valarray<double> dx, int l, int m);
  double dist_par_z(double x, double y);  // get distance from line (x,y) || z to closest atom

  // --- neighbour list

  void init_nlist(double cut_pot, double cut_list);  // initialize neighbour list
  void update_linkedlist();  // update linked list
  void add_to_linkedlist(int i);  // add an atom to linked list
  void update_nlist();
  bool update_already();
  void check_nlist();
  void update_symmnlist();
};
#endif
