/*
  ------------------------------------------------------------------------------

    Copyright © 2005-2012, Tommi T. Järvi
                           firstname . jarvi @ iki . fi (remove spaces)


    This file is part of one of my programs.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2,
    as published by the Free Software Foundation.

    This is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  ------------------------------------------------------------------------------
*/


#include <valarray>
#include <iostream>
#include <vector>
#include <map>
#include "global.h"
#include "atomsystem.h"

using namespace std;


class CNA_Number {

  /*
    A bond signature for common neighbour analysis
  */


 public:
  int nn;  // 1 if atoms are neighbours, 0 otherwise
  int n;  // number of common neighbours
  int b;  // bonds between common neighbours
  int t;  // topology

  CNA_Number() {nn=-1; n=-1; b=-1; t=-1;}  // default constructor
  CNA_Number(int n1, int n2, int n3, int n4) {nn=n1; n=n2; b=n3; t=n4;}  // constructor

  CNA_Number& operator=(const CNA_Number &other);
  bool operator==(const CNA_Number &other) const {
    if(nn==other.nn && n==other.n && b==other.b && t==other.t) return true;
    else return false;
  }
  bool operator<(const CNA_Number &other) const;

  void reset() {nn=-1; n=0; b=0; t=-1;}  // reset
};


class CNA_Signature {

  /*
    Atom signature for common neighbour analysis

    At the moment, the maximum number of bonds per atom is fixed.
  */


 public:
  int nbonds;
  map <CNA_Number, int> bonds;  // bonds (signature, count)

  CNA_Signature() {nbonds = 0;}  // default constructor

  CNA_Signature& operator=(const CNA_Signature &other) {nbonds=other.nbonds; bonds=other.bonds; return *this;}
  bool operator==(const CNA_Signature &other) const { if(bonds==other.bonds && nbonds==other.nbonds) return true; else return false;}
  bool operator<(const CNA_Signature &other) const {if(bonds<other.bonds) return true; else return false;}

  void add(CNA_Number bond) {bonds[bond]++; nbonds++;}
  void reset() {bonds.clear(); nbonds=0;}
};


class CNA {

  /*
    Common neighbour analysis.


    mode          explanation
    -------------------------
    0             normal mode (not implemented)
    1             only atom pairs with bonds between them
    2             1 + mark surface atoms

    Note: The id number in "type" can be any integer (two types can
          even have the same id). Unknowns are marked with 0 in any
          case. For mode 2, surface atoms are marked with a negative
          number that is not used already (usually -1).
  */


 public:

  vector<CNA_Signature> signatures;        // CNA-signatures, [0,natoms-1]
  map<CNA_Signature, int> unknowns_count;  // unidentified signatures, count
  map<CNA_Signature, int> unknowns_type;   // unidentified signatures, type given
  double bondcut;                          // bond cutoff
  int mode;                                // 0 - normal, 1 - only pairs with bond between the atoms
  map<CNA_Number, int> stats;              // statistics of bond signatures
  bool track_unknowns;                     // keep track of unknown sig's, need to turn on externally to output new sigs with savesigs()

  vector<CNA_Signature> data;              // known signatures
  vector<string> name;                     // explanation for each sig.
  vector<int> type;                        // number -''-

  double surf_coord;                       // mode 2: if coordination < than this, count as surface atom

  // default constructor
  CNA() {
    bondcut = -1;
    mode = 1;
    track_unknowns = false;
    surf_coord = 10;
  }

  void print_unknowns();
  void print_stats(double time = 0.0);
  void add_unknown(const CNA_Signature &sig, int &newtype);
  void calculate_and_analyze(Atom_System &s, double time = 0);
  void reset() {signatures.clear(); unknowns_count.clear(); unknowns_type.clear(); bondcut=-1; mode=1; track_unknowns = false; stats.clear();}
  void savesigs(string file);
  void readsigs(string file);
};
